<?php
// Own namespace
namespace CoreFramework\Index\Stack;

// Import framework stuff
use CoreFramework\Index\BaseIndex;
use CoreFramework\Index\Indexable;
use CoreFramework\Registry\Registerable;
use CoreFramework\Stack\Index\IndexableStack;

/**
 * A FileStack index class
 *
 * @author		Roland Haeder <webmaster@ship-simu.org>
 * @version		0.0.0
 * @copyright	Copyright (c) 2007, 2008 Roland Haeder, 2009 - 2017 Core Developer Team
 * @license		GNU GPL 3.0 or any newer version
 * @link		http://www.ship-simu.org
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 */
class FileStackIndex extends BaseIndex implements IndexableStack, Registerable {
	/**
	 * Protected constructor
	 *
	 * @return	void
	 */
	protected function __construct () {
		// Call parent constructor
		parent::__construct(__CLASS__);
	}

	/**
	 * Creates an instance of this Index class and prepares it for usage
	 *
	 * @param	$fileName		Name of the file stack to create an index file for
	 * @return	$indexInstance	An instance of this Index class
	 */
	public final static function createFileStackIndex ($fileName) {
		// Get a new instance
		$indexInstance = new FileStackIndex();

		// Initialize index
		$indexInstance->initIndex($fileName);

		// Return the prepared instance
		return $indexInstance;
	}

	/**
	 * Adds given hash to an index file
	 *
	 * @param	$groupId	Name of stack to add hash for
	 * @param	$data		Hash and gap position to be added to the index
	 * @return	void
	 */
	public function addHashToIndex ($groupId, array $data) {
		// Debug message
		/* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput(sprintf('[%s:%d:] groupId=%s,hash=%s,gap=%d,length=%d - CALLED!', __METHOD__, __LINE__, $groupId, $data[BaseFileStack::ARRAY_INDEX_HASH], $data[BaseFileStack::ARRAY_INDEX_GAP_POSITION], $data[BaseFileStack::ARRAY_INDEX_DATA_LENGTH]));

		// Raw data been written to the file
		$rawData = sprintf('%s%s%s%s%s%s%s',
			$groupId,
			self::SEPARATOR_GROUP_HASH,
			hex2bin($data[BaseFileStack::ARRAY_INDEX_HASH]),
			self::SEPARATOR_HASH_GAP_POSITION,
			$data[BaseFileStack::ARRAY_INDEX_GAP_POSITION],
			self::SEPARATOR_GAP_LENGTH,
			$data[BaseFileStack::ARRAY_INDEX_DATA_LENGTH]
		);

		// Debug message
		/* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput(sprintf('[%s:%d:] groupId=%s,hash=%s,rawData()=%d', __METHOD__, __LINE__, $groupId, $data[BaseFileStack::ARRAY_INDEX_HASH], strlen($rawData)));

		// Search for next free gap
		$gapPosition = $this->getIteratorInstance()->searchNextGap(strlen($rawData));

		// Gap position cannot be smaller than header length + 1
		assert($gapPosition > $this->getIteratorInstance()->getHeaderSize());

		// Debug message
		/* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput(sprintf('[%s:%d:] groupId=%s,hash=%s,gapPosition=%s', __METHOD__, __LINE__, $groupId, $data[BaseFileStack::ARRAY_INDEX_HASH], $gapPosition));

		// Then write the data at that gap
		$this->getIteratorInstance()->writeData($gapPosition, $rawData);

		// Debug message
		/* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput(sprintf('[%s:%d:] groupId=%s,hash=%s,rawData()=%d - EXIT!', __METHOD__, __LINE__, $groupId, $data[BaseFileStack::ARRAY_INDEX_HASH], strlen($rawData)));
	}

	/**
	 * Searches for next suitable gap the given length of data can fit in
	 * including padding bytes.
	 *
	 * @param	$length			Length of raw data
	 * @return	$seekPosition	Found next gap's seek position
	 */
	public function searchNextGap ($length) {
		$this->partialStub('length=' . $length);
	}

}
